import os
import re
import sys
import time
import logging
import argparse
import pandas as pd
import numpy as np


'''
Description:
This module is used for estimating TMRCA of haplotypes.
Network information file or sequence alignment file is supported to estimate TMRCA.
The program will calculate rho statistics from network information file, or calculate ASD statistics from the sequence alignment file.
Then estimate TMRCA according to statistics.
'''


def tmrca_parser():

    parser = argparse.ArgumentParser('tmrca', description='(c) Y-LineageTracker: TMRCA estimation')
    # function used for TMRCA estimation
    parser.add_argument('tmrca',
                        help='Estimate TMRCA of Y haplotypes.')
    # required, network information file or matrix file
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--net',
                        required=False,
                        type=str,
                        action='store',
                        help='net: Network information file.')
    group.add_argument('--matrix',
                        required=False,
                        type=str,
                        action='store',
                        help='matrix: Y-STR haplotype data in matrix format')
    # required, information of ancestor and descendants.
    parser.add_argument('--anc-des',
                        required=True,
                        type=str,
                        dest='ad',
                        action='append',
                        help='anc-des: Information of ancestor and descendants. \
                              The program will estimate the time based on mutations from the ancestor to each descendant.')
    # required, population file
    parser.add_argument('-p', '--population',
                        required=False,
                        type=str,
                        action='store',
                        help='population: A file containing sample ID and population information of each individual.')
    # optional, mutation rate of Y-STR
    parser.add_argument('--mut-rate',
                        required=False,
                        type=float,
                        dest='rate',
                        action='store',
                        default=6e-3,
                        help='mut-rate: The general mutation rate across Y-STR haplotype.')
    # optional, generation time
    parser.add_argument('--generation',
                        required=False,
                        type=int,
                        action='store',
                        default=30,
                        help='generation: The generation time used for time estimation.')
    # optional, the prefix of output
    parser.add_argument('-o', '--output',
                        required=False,
                        type=str,
                        action='store',
                        help='output: The prefix of output files.')

    args = parser.parse_args()

    return args


# print program information and write to log file
def set_log(log_file, args_log):

    logger = logging.getLogger()
    logger.setLevel(level=logging.INFO)

    handler = logging.FileHandler(log_file, mode='w')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] - [%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    logger.addHandler(handler)
    logger.addHandler(console)

    log_info = ['[Y-LineageTracker] [TMRCA]',
                '[Y-LineageTracker] Run Date: ' + time.asctime(time.localtime(time.time()))]

    if args_log.net:
        log_info.append('[Y-LineageTracker] Network Info File: %s' % args_log.net)
        stat = 'Rho'
    elif args_log.matrix:
        log_info.append('[Y-LineageTracker] Matrix File: %s' % args_log.matrix)
        stat = 'ASD'

    log_info.extend(['[Y-LineageTracker] Ancestor and Descendants Information: %s' % args_log.ad,
                     '[Y-LineageTracker] Statistics: %s' % stat,
                     '[Y-LineageTracker] Mutation Rate: %s' % args_log.rate,
                     '[Y-LineageTracker] Generation: %s' % args_log.generation])

    print('\n')
    for i in log_info:
        logger.info(i)


class CommonAncestor(object):
    '''
    This class is used to estimate TMRCA of haplotypes
    The program can calculate:
    1. rho statistics from network information file
    2. ASD statistics from matrix file
    There is only one output, that is the TMRCA of haplotypes
    Note that the time estimated by ASD is larger than that of rho
    '''

    def __init__(self):

        from GetConfig import getConfig
        config = getConfig()
        self.logger = logging.getLogger()
        self.skip_letters = config.get('Statistics', 'SkipLetters').split(',')

    def _parse_hap_data(self, matrix_file, population_file):

        from FilesIO import check_hap_input, check_population, check_pop_header, check_overlap, set_haplotype_index
        hap_data = check_hap_input(matrix_file, 'haplotype', 'matrix')

        population_data = check_population(population_file)
        population_data = check_pop_header(population_data)

        hap_data, population_data = check_overlap(hap_data, population_data)
        hap_data = set_haplotype_index(hap_data, 'matrix')
        population_data = population_data.set_index(['SampleID'], drop=True)

        return hap_data, population_data

    def _parse_net_data(self, net_file):

        net = open(net_file, 'r').readlines()

        for i, j in enumerate(net):
            if j.startswith('#HapLength'):
                length = int(j.strip().split(':')[-1])
            elif j.startswith('#HapSamples'):
                sample_line = i
            elif j.startswith('#HapLinks'):
                link_line = i

        sample_info = net[sample_line+1:link_line-1]
        sample_num = {}
        for i in sample_info:
            line = i.strip()
            num = int(line.split('\t')[0])
            samples = line.split('\t')[-1].split(',')
            for j in samples:
                sample_num[j] = num

        link_info = net[link_line+1:]
        link_data = np.array([i.strip().split('\t') for i in link_info])
        link_data = link_data.astype(int)

        net_data = [length, sample_num, link_data]

        return net_data

    def _get_ancestor_and_descendants(self, anc_des_list):

        if len(anc_des_list) == 1:
            anc_des = anc_des_list[0]
            if os.path.isfile(anc_des):
                pre_ad_list = open(anc_des).readlines()
            else:
                pre_ad_list = [anc_des]
        else:
            pre_ad_list = anc_des_list

        ad_list = []
        for i in pre_ad_list:
            if re.match(r'^\w+:(\w+,)+\w+$', i):
                ad_list.append(i)
            else:
                self.logger.error('[Y-LineageTracker] %s format error' % i)
                sys.exit()

        return ad_list

    def _calculate_tmrca(self, data, rate, generation, ad_list, stat, path, population_data):

        if stat == 'rho':
            L = data[0]
            sample_num = data[1]
            link_data = data[2]
        elif stat == 'asd':
            L = data.columns.size # haplotype length
        mu = rate # mutation rate
        g = generation # generation

        num = 0
        out_df = pd.DataFrame(columns=['Ancestor', 'Descendants', stat, 'TMRCA'])

        for ad in ad_list:
            a = ad.split(':')[0]
            d = ad.split(':')[1].split(',')

            if stat == 'rho':
                ancestor_node = sample_num[a]
                descendant_nodes = [sample_num[i] for i in d]
                res = self._calculate_rho(ancestor_node, descendant_nodes, link_data)
            elif stat == 'asd':
                pop_inds = {}
                for pop in sorted(set(population_data['Population'])):
                    inds = population_data[population_data['Population']==pop].index.tolist()
                    pop_data = data.loc[inds].to_numpy().T
                    pop_inds[pop] = pop_data

                ancestor_hap = data.loc[a, :].tolist()
                ancestor_pop = population_data.at[a, 'Population']
                ancestor_data = pop_inds[ancestor_pop]
                ancestor_freq = [list(j).count(i) / len(list(filter(lambda x: x not in self.skip_letters, j))) for i, j in zip(ancestor_hap, ancestor_data)]

                descendant_haps = [data.loc[i, :].tolist() for i in d]
                descendant_freqs = []
                for d_ind, d_hap in zip(d, descendant_haps):
                    descendant_pop = population_data.at[d_ind, 'Population']
                    descendant_data = pop_inds[descendant_pop]
                    descendant_freq = [list(j).count(i) / len(list(filter(lambda x: x not in self.skip_letters, j))) for i, j in zip(d_hap, descendant_data)]
                    descendant_freqs.append(descendant_freq)

                res = self._calculate_asd(ancestor_hap, descendant_haps, ancestor_freq, descendant_freqs)
            time = round((res / (mu*L)) * g, 6)
            line = [a, ','.join(d), res, time]
            out_df.loc[num] = line
            num += 1

        out_df.to_csv(path + '.%s.txt' % stat, sep='\t', index=False)

    # the average number of sites differing between a set of haplotypes and their specified common ancestor
    def _calculate_asd(self, ancestor_hap, descendant_haps, ancestor_freq, descendant_freqs):

        diffs = []
        for d_hap, d_freq in zip(descendant_haps, descendant_freqs):
            diff_num = np.sum([(int(i)-int(j))**2*k*l for i, j, k, l in zip(ancestor_hap, d_hap, ancestor_freq, d_freq) if i not in self.skip_letters and j not in self.skip_letters])
            diffs.append(diff_num)

        res = round(np.sum(diffs), 2)

        return res

    def _calculate_rho(self, ancestor_node, descendant_nodes, link_data):

        end = ancestor_node
        diff_num = []
        for d_node in descendant_nodes:
            start = d_node
            diff = self._calculate_net_rho(start, end, link_data)
            diff_num.append(diff)

        res = round(np.mean(diff_num), 2)

        return res

    def _calculate_net_rho(self, start, end, link_data):

        vs = [start]
        num = 0
        while True:
            if num == 0:
                vs2 = [start]
                steps = [0]
            else:
                vs2 = vs3
                steps = steps2
            vs3 = []
            steps2 = []
            for i, j in zip(vs2, steps):
                v = link_data[np.where(link_data[:, :2]==i)[0]]
                nodes = sorted(list(set(v[:,:2].reshape(-1))))
                nodes2 = list(filter(lambda x: not x in vs, nodes))
                stp = (v[:,2]+j).tolist()
                stp = [stp[np.where(v==n)[0][0]] for n in nodes2]
                vs3.extend(nodes2)
                steps2.extend(stp)
                vs.extend(vs3)
            num += 1
            if end in vs3:
                diff_num = min(steps2)
                break

        return diff_num

    def run_tmrca(self, arguments, path):

        if arguments.net:
            data = self._parse_net_data(arguments.net)
            stat = 'rho'
            population_data = None
        elif arguments.matrix:
            if arguments.population:
                data, population_data = self._parse_hap_data(arguments.matrix, arguments.population)
                stat = 'asd'
            else:
                self.logger.error('[Y-LineageTracker] Population file is required for matrix file to perform ASD estimation')
                sys.exit()

        self.logger.info('[Y-LineageTracker] Reading input data')
        ad_list = self._get_ancestor_and_descendants(arguments.ad)
        self.logger.info('[Y-LineageTracker] Get ancestor and descendants')
        self.logger.info('[Y-LineageTracker] Calculating TMRCA')
        self._calculate_tmrca(data, arguments.rate, arguments.generation, ad_list, stat, path, population_data)
        self.logger.info('[Y-LineageTracker] Finished')


def main():

    from FilesIO import get_out_path, time_count
    start = time.perf_counter()
    arguments = tmrca_parser()
    path = get_out_path(arguments.output, 'TMRCA')
    log_file = path + '.TMRCALog.log'

    set_log(log_file, arguments)

    estimate = CommonAncestor()
    estimate.run_tmrca(arguments, path)

    time_count(start)

if __name__ == '__main__':
    main()



import os
import sys
import gzip
import tempfile
import numpy as np
import pandas as pd

sys.path.append(os.path.split(os.path.realpath(__file__))[0]+'/..')
from Bio import AlignIO, SeqIO, Phylo
from GetConfig import getConfig


config = getConfig()


class ConvertData(object):
    '''
    This class is mainly used to convert different format of data
    Conversion of data is helpful for some analysis
    This class can:
    1. convert VCF to MSA and matrix
    2. convert matrix to MSA
    3. convert MSA to matrix
    4. convert MSA in different formats
    MSA includes: fasta, phylip, nexus, mega
    VCF: variant call format
    MSA: multiple sequence alignment
    '''

    def __init__(self, in_file):

        from FilesIO import  check_compress
        self.vcf_header_cut = config.getint('DataInfo', 'VcfHeaderCut')
        self.compressed = check_compress(in_file)
        self.in_file = in_file

    # convert VCF to sequence format
    def _match_to_vcf(self, file, out_format):

        # dictionary of IUPAC ambiguities for nucleotides
        # '*' means deletion for GATK
        # deletions are ignored when making the consensus
        amb = {('A','A'):'A',
               ('A','N'):'A',
               ('C','C'):'C',
               ('C','N'):'C',
               ('G','G'):'G',
               ('G','N'):'G',
               ('N','A'):'A',
               ('N','C'):'C',
               ('N','G'):'G',
               ('N','N'):'N',
               ('N','T'):'T',
               ('T','N'):'T',
               ('T','T'):'T',
               ('*','*'):'-',
               ('A','*'):'A',
               ('*','A'):'A',
               ('C','*'):'C',
               ('*','C'):'C',
               ('G','*'):'G',
               ('*','G'):'G',
               ('T','*'):'T',
               ('*','T'):'T',
               ('N','*'):'N',
               ('*','N'):'N'}

        if self.compressed:
            vcf = gzip.open(file, 'rt')
        else:
            vcf = open(file, 'r')
        with vcf:

            # create a list to store sample names
            sample_names = []

            # keep track of longest sequence name for padding with spaces in the output file
            len_longest_name = 0

            # look for the line in the VCF header with the sample names
            for line in vcf:
                if line.startswith('#CHROM'):

                    # split line into fields
                    broken = line.strip('\n').split('\t')
                    min_samples_locus = int(len(broken[self.vcf_header_cut:])*0.8)

                # Create a list of sample names and the keep track of the longest name length
                    for i in range(self.vcf_header_cut, len(broken)):
                        name_sample = broken[i].replace('./', '') # GATK adds './' to sample names sometimes
                        sample_names.append(name_sample)
                        len_longest_name = max(len_longest_name, len(name_sample))

                # Find out the ploidy of the genotypes
                elif not line.startswith('#'):
                    break
        vcf.close()

        # create an intermediate file to hold the sequence data vertically and then transpose it to create the matrices
        if out_format == 'matrix':
            tmp_list = []
        else:
            tmp_file = tempfile.NamedTemporaryFile(mode='w')
        # process vcf
        index_last_sample = len(sample_names) + self.vcf_header_cut
        # initialize line counter
        snp_num = 0
        snp_accepted = 0
        snp_shallow = 0
        snp_multinuc = 0
        snp_biallelic = 0
        ploidy = 1
        if self.compressed:
            vcf = gzip.open(file, 'rt')
        else:
            vcf = open(file, 'r')
        with vcf:
            while True:
                # load large chunks of file into memory
                vcf_chunk = vcf.readlines(100000)
                if not vcf_chunk:
                    break
                # process the SNPs one by one
                for line in vcf_chunk:
                    if not line.startswith('#') and line.strip('\n') != '': # pyrad sometimes produces an empty line after the #CHROM line
                        # split line into columns
                        broken = line.strip('\n').split('\t')
                        for g in range(self.vcf_header_cut, len(broken)):
                            if broken[g] in ['.', '.|.']:
                                broken[g] = './.'
                        # keep track of number of genotypes processed
                        snp_num += 1
                        # check if the SNP has the minimum of samples required
                        if (len(broken[self.vcf_header_cut:]) - ''.join(broken[self.vcf_header_cut:]).count('./.')) >= min_samples_locus:
                            # check that ref genotype is a single nucleotide and alternative genotypes are single nucleotides
                            if len(broken[3]) == 1 and (len(broken[4])-broken[4].count(',')) == (broken[4].count(',')+1):
                                # add to running sum of accepted SNPs
                                snp_accepted += 1
                                # create a dictionary for genotype to nucleotide translation
                                # each SNP may code the nucleotides in a different manner
                                nuc = {str(0): broken[3], '.': 'N'}
                                for n in range(len(broken[4].split(','))):
                                    nuc[str(n+1)] = broken[4].split(',')[n]
                                # translate genotypes into nucleotides and the obtain the IUPAC ambiguity
                                # for heterozygous SNPs, and append to DNA sequence of each sample
                                if ploidy == 1:
                                    site_list = [(amb[(nuc[broken[i][0]], nuc[broken[i][0]])]) for i in range(self.vcf_header_cut, index_last_sample)]
                                    site_tmp = ''.join(site_list)
                                # write entire row of single nucleotide genotypes to temporary file of fasta and nexus
                                if out_format == 'matrix':
                                    test_list.append(site_list)
                                else:
                                    tmp_file.write(site_tmp+'\n')
                            else:
                                # keep track of loci rejected due to multinucleotide genotypes
                                snp_multinuc += 1
                                # keep track of loci rejected due to exceeded missing data
                                snp_shallow += 1
                        else:
                            # keep track of loci rejected due to exceeded missing data
                            snp_shallow += 1

        if out_format == 'matrix':
            matrix = np.array(test_list).T
            matrix_df = pd.DataFrame(matrix, index=sample_names, columns=['V'+str(i) for i in range(np.size(test2, 1))])
            matrix_df.index.name = 'SampleID'
        else:
            return tmp_file, sample_names, len_longest_name, snp_accepted

    # convert VCF to sequence format
    def _write_matched_tmp_to_seq(self, tmp_file, sample_names, len_longest_name, snp_accepted, out_format):

        tmp_file2 = tempfile.NamedTemporaryFile(mode='w')

        # output file type for sebsequent analysis
        if out_format == 'phylip':
            header_phy = str(len(sample_names))+' '+str(snp_accepted)+'\n'
            tmp_file2.write(header_phy)
        elif out_format == 'nexus':
            header_nex = '#NEXUS\n\nBEGIN DATA;\n\tDIMENSIONS NTAX=' + str(len(sample_names)) + ' NCHAR=' + str(snp_accepted) + ';\n\tFORMAT DATATYPE=DNA' + ' MISSING=N' + ' GAP=- ;\nMATRIX\n'
            tmp_file2.write(header_nex)

        # Write sequences of the ingroup
        for s in range(0, len(sample_names)):
            with open(tmp_file.name) as tmp_seq:
                seqout = ''
                # where the transposing happens
                for line in tmp_seq:
                    seqout += line[s]
                if out_format == 'fasta':
                    # write FASTA line
                    tmp_file2.write('>'+sample_names[s]+'\n'+seqout+'\n')
                else:
                    # pad sequences names and write PHYLIP or NEXUS lines
                    padding = (len_longest_name+3-len(sample_names[s])) * ' '
                    if out_format == 'phylip':
                        tmp_file2.write(sample_names[s]+padding+seqout+'\n')
                    if out_format == 'nexus':
                        tmp_file2.write(sample_names[s]+padding+seqout+'\n')

        res = AlignIO.read(open(tmp_file2.name), out_format)

        return res

    # convert VCF to msa
    def _vcf_to_msa(self, file, out_format):

        tmp_file, sample_names, len_longest_name, snp_accepted = self._match_to_vcf(file, out_format)
        res = self._write_matched_tmp_to_seq(tmp_file, sample_names, len_longest_name, snp_accepted, out_format)

        return res

    # convert VCF to matrix format
    def _vcf_to_matrix(self, file, out_format='matrix'):

        matrix_df = self._match_to_vcf(file, out_format)

        return matrix_df

    # convert msa to matrix
    def _seq_to_matrix(self, file, in_format, tmp):

        if tmp:
            msa = AlignIO.read(open(file.name), in_format)
        else:
            msa = AlignIO.read(open(file), in_format)
        ids = [i.id for i in msa]
        seqs = [i.seq for i in msa]
        seq_len = len(seqs[0])
        seq_matrix = np.array([[j]+[x for x in i] for i, j in zip(seqs, ids)])

        matirx = pd.DataFrame(columns=['SampleID']+['V'+str(i) for i in range(seq_len)], data=seq_matrix)

        return matirx

    # read seq to msa
    def _seq_to_msa(self, file, in_format, tmp):

        if tmp:
            res = AlignIO.read(open(file.name), in_format)
        else:
            res = AlignIO.read(open(file), in_format)

        return res

    # convert msa in different
    def _seq_convert(self, file, in_format, out_format, tmp):

        tmp_file = tempfile.NamedTemporaryFile(mode='w')
        if tmp:
            SeqIO.convert(file.name, in_format+'-relaxed', tmp_file, out_format)
        else:
            SeqIO.convert(file, in_format+'-relaxed', tmp_file, out_format)
        res = AlignIO.read(open(tmp_file.name), out_format)

        return res

    # convert matrix to msa
    def _matrix_to_msa(self, file, out_format):

        matrix_info = np.loadtxt(file, dtype='object').T
        snp_accepted = len(matrix_info[:,0]) - 1
        sample_names= matrix_info[0,:].tolist()[1:]
        tmp_file = tempfile.NamedTemporaryFile(mode='w')
        for i in matrix_info[1:, 1:]:
            tmp_file.write(''.join(i)+'\n')

        res = self._write_matched_tmp_to_seq(tmp_file, sample_names, snp_accepted, out_format)

        return res

    # convert meg to fasta
    def _meg_to_fasta(self, file):

        meg = open(file, 'r').readlines()
        tmp_file = tempfile.NamedTemporaryFile(mode='w')

        with open(tmp_file.name, 'w') as tmp:
            for i in meg[3:]:
                if i.startswith('#'):
                    tmp.write(i.replace('#', '>', 1))
                else:
                    tmp.write(i)

        return tmp_file

    # main function of data conversion
    def data_convert(self, in_format, out_format):

        seq_format = ['fasta', 'phylip', 'nexus']
        in_file = self.in_file

        if in_format == 'vcf':
            if out_format == 'msa':
                res = self._vcf_to_msa(in_file, 'fasta')
            elif out_format == 'matrix':
                res = self._vcf_to_martix(in_file)
        elif in_format == 'matrix':
            res = self._matrix_to_msa(in_file, out_format)
        else:
            if in_format == 'meg':
                in_file = self._meg_to_fasta(in_file)
                in_format = 'fasta'
                tmp = True
            else:
                tmp = False
            if out_format == 'matrix':
                res = self._seq_to_matrix(in_file, in_format, tmp)
            elif out_format == 'msa':
                res = self._seq_to_msa(in_file, in_format, tmp)
            else:
                res = self._seq_convert(in_file, in_format, out_format, tmp)

        return res



import os
import sys
import time
import logging
import argparse
import itertools
import matplotlib; matplotlib.use('Agg')
import multiprocessing
import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

from operator import add
from functools import partial
from collections import Counter
from matplotlib import font_manager
from GetConfig import getConfig


config = getConfig()


'''
Description:
This module is used for network analysis.
Network analysis is an efficient method to show relationship of haplotypes.
Y haplotype data is required for analysis.
It is recommended to use Y-STR haplotype under a specific haplogroup.
Network analysis can show the evolutionary relationship between haplotypes.
This command can output network figure, and fdi file.
'''


def net_parser():

    parser = argparse.ArgumentParser('net', description='(c) Y-LineageTracker: Network analysis')
    # function used for network analysis
    parser.add_argument('net',
                        help='Perform network analysis from Y haplotype data.')
    # required, input file, alignment sequence or matrix
    input = parser.add_mutually_exclusive_group(required=True)
    input.add_argument('--seq',
                        type=str,
                        action='store',
                        help='seq: Haplotype data in sequence alignment format')
    input.add_argument('--matrix',
                        type=str,
                        action='store',
                        help='martix: Y-STR haplotype data in matrix format')
    # optional, format of sequence alignment
    parser.add_argument('--seq-format',
                        required=False,
                        type=str,
                        dest='format',
                        action='store',
                        choices=['fasta', 'phylip', 'nexus', 'meg', 'vcf'],
                        help='--seq-format: The format of sequence file. This option is only required for the sequence alignment file. ')
    # optional, population file
    parser.add_argument('-p', '--population',
                        required=False,
                        type=str,
                        action='store',
                        help='population: A file containing sample ID and population information of each individual.')
    # optional, type of tree
    parser.add_argument('--tree-type',
                        required=False,
                        type=str,
                        dest='treetype',
                        action='store',
                        default='mjn',
                        choices=['mjn', 'msn'],
                        help='tree-type: The algorithm used to construct a network tree.')
    # optional, how to code missing site
    parser.add_argument('--gap',
                        required=False,
                        type=str,
                        action='store',
                        default='allele',
                        choices=['allele', 'missing'],
                        help='gap: The gap coding method. You can choose to code gaps as a new type of allele or missing data')
    # optional, the search for new median nodes
    parser.add_argument('--search',
                        required=False,
                        type=int,
                        action='store',
                        default=0,
                        help='search: How the search for new median nodes is performed: the larger this parameter, the wider the search. ')
    # optional, output fdi file
    parser.add_argument('--fdi',
                        required=False,
                        action='store_true',
                        help='fdi: Output fdi file that can be used as the input file in Network software, to show network plot or estimate TMRCA directly.')
    # optional, filter sites by misisng rate value
    parser.add_argument('--filter',
                        required=False,
                        type=float,
                        action='store',
                        default=0.7,
                        help='filter: The cutoff value of the site to be filtered, sites with a missing rate value greater than this value will be filtered')
    # optional, how to show alternative link in figure
    parser.add_argument('--alt',
                        required=False,
                        type=str,
                        action='store',
                        default='hide',
                        choices=['hide', 'show', 'dashed'],
                        help='alt: Whether to show alternative links between nodes in the network figure.')
    # optional, size of median nodes
    parser.add_argument('--mediansize',
                        required=False,
                        type=float,
                        action='store',
                        default=0.0,
                        help='mediansize: The median node size in the median-joining network compared to base size in the network figure. The default is 0.')
    # optional, keep all nodes in same size
    parser.add_argument('--euqal-size',
                        required=False,
                        dest='euqalsize',
                        action='store_true',
                        help='euqal-size: Keep all nodes the same size in the network figure.')
    # optional, the prefix of output
    parser.add_argument('-o', '--output',
                        required=False,
                        type=str,
                        action='store',
                        help='output: The prefix of output files. ')

    args = parser.parse_args()

    return args


# check missing rate is filter argument is used
def check_misisng_rate(missing_rate):

    if missing_rate < 0 or missing_rate> 1:
        print('[Y-LineageTracker] [Error] Cutoff of missing rate should be a float between 0 and 1')
        sys.exit()


# print program information and write to log file
def set_log(args_log, log_file):

    logger = logging.getLogger()
    logger.setLevel(level=logging.INFO)

    handler = logging.FileHandler(log_file, mode='w')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] - [%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    logger.addHandler(handler)
    logger.addHandler(console)

    log_info = ['[Y-LineageTracker] [Network]',
                '[Y-LineageTracker] Run Date: ' + time.asctime(time.localtime(time.time())),
                '[Y-LineageTracker] Haplotype File: %s' % args_log.seq,
                '[Y-LineageTracker] Haplotype File Format: %s' % args_log.format,
                '[Y-LineageTracker] Gap euqal to %s' % args_log.gap,
                '[Y-LineageTracker] Searching parameter: %d' % args_log.search,
                '[Y-LineageTracker] Missing rate of allele to be filtered: %s' % args_log.filter,
                '[Y-LineageTracker] Alternative link type : %s' % args_log.alt,
                '[Y-LineageTracker] Median node size: %s' % args_log.mediansize]

    if args_log.euqalsize:
        log_info.append('[Y-LineageTracker] Keep all node the same size')
    else:
        log_info.append('[Y-LineageTracker] Plot node according to sample size')

    print('\n')
    for i in log_info:
        logger.info(i)


class HaploNetwork(object):
    '''
    This class is used for network analysis.
    Main functions of this class include:
    1. construct median-joining network from haplotypes
    2. plot network by networkx package
    3. write network to fdi file
    4. output network info
    '''

    def __init__(self, gap, path):

        self.logger = logging.getLogger()
        self.skip_letters = config.get('Statistics', 'SkipLetters').split(',')
        self.gap = gap
        self.path = path
        self.net_info = open(self.path+'.info.net', 'w')

    # convert STR numbers to 0-1 code haplotype
    def _convert_allel_to_binary(self, haplotype_data, filter):

        # create a empty dataframe to save haplotypes
        sequence_df = pd.DataFrame(index=haplotype_data.index,
                                   columns=['Haplotype'])
        sequence_df['Haplotype'] = ''

        STR_code_pos = {}
        start_pos = 0
        hap_length = 0
        allel_num = haplotype_data.index.size
        for STR in haplotype_data.columns.tolist():
            str_allels = haplotype_data[STR].astype(str).tolist()
            # filter STR according to filter parameter
            if sum([str_allels.count(i) for i in self.skip_letters])/allel_num >= filter:
                continue
            if len(set(set(str_allels)&set(self.skip_letters))) > 1:
                str_allels = ['.' if i in self.skip_letters else i for i in str_allels]
            repeats_type = sorted(list(set(str_allels)))
            # convert STR repeat numbers to 0-1 code
            repeats_type_num = len(repeats_type)
            max_length = len('{0:b}'.format(repeats_type_num-1))
            binary_set = [] # creat a empty list to save 0-1 code
            for i in range(repeats_type_num):
                binary_num = '{0:b}'.format(i)
                binary_num_length = len(binary_num)
                if binary_num_length < max_length:
                    binary_num = (max_length-binary_num_length)*'0'+binary_num
                binary_set.append(binary_num)
            if self.gap == 'missing':
                replaced_allels = ['-'*max_length if i == '.' else binary_set[repeats_type.index(i)] for i in str_allels]
            else:
                replaced_allels = [binary_set[repeats_type.index(i)] for i in str_allels]
            sequence_df['Haplotype'] = list(map(add, sequence_df['Haplotype'].tolist(), replaced_allels))
            for pos in range(start_pos, start_pos+max_length):
                STR_code_pos[pos] = STR
            start_pos += max_length
            hap_length += 1

        self.net_info.write('#HapLength:%d\n\n' % hap_length)

        return sequence_df, STR_code_pos

    # calculate frequency of all kinds of haplotypes and the proportion in each population
    def _calculate_distance_matrix(self, sequence_df):

        # get haplotypes and name for them
        haplotypes = sorted(list(set(sequence_df['Haplotype'])))
        sp_sequence_df = pd.DataFrame(index=range(len(haplotypes)), columns=['Haplotype'])
        sp_sequence_df['Haplotype'] = haplotypes

        # calculate distance between two haplotypes
        distance_matrix = pd.DataFrame(index=range(len(haplotypes)), columns=range(len(haplotypes)))
        hap_num = 0
        for hap1_name, hap1 in enumerate(haplotypes):
            hap_num += 1
            for hap2_name, hap2 in enumerate(haplotypes[hap_num:]):
                if self.gap == 'missing':
                    distance = sum(l1 != l2 for l1, l2 in zip(hap1, hap2) if l1 not in self.skip_letters and l2 not in self.skip_letters)
                    if distance == 0:
                        distance = 1
                else:
                    distance = sum(l1 != l2 for l1, l2 in zip(hap1, hap2))
                hap2_name = hap2_name + hap_num
                distance_matrix.at[hap2_name, hap1_name] = distance

        self.net_info.write('#HapSamples:\n')

        sample_sp_code = {}
        for i in haplotypes:
            idx = sequence_df['Haplotype'].tolist().index(i)
            sample = sequence_df.index[idx]
            sample_sp_code[sample] = i
            all_samples = ','.join(sequence_df[sequence_df['Haplotype']==i].index.tolist())
            self.net_info.write('%d\t%s\n' %(idx, all_samples))

        self.net_info.write('\n')

        return distance_matrix, sp_sequence_df, sample_sp_code

    # get msn tree
    def _calculate_minimum_spanding_tree(self, n, distance_matrix, return_pruned=True):

        if n < 2:
            self.logger.error('[Y-LineageTracker] Program stopped since there are less than 2 types of haplotypes')
            sys.exit()

        # create a empty edge matrix to save hapotype and their distance
        edge_num = n-1
        edge_matrix = np.empty(shape=(edge_num, 3))
        edge_matrix[:] = np.nan

        # get the distance from small to large between of two haplotypes
        forest = np.linspace(0, n-1, n)

        add_num = 0
        while add_num < edge_num:
            ci, ri = np.unravel_index(np.nanargmin(distance_matrix.T.values), distance_matrix.T.shape)
            row_record = forest[ri]
            col_record = forest[ci]
            if row_record != col_record:
                edge_matrix[add_num] = [ci, ri, distance_matrix.at[ri, ci]] # col hap, row hap, step
                forest[forest==col_record] = row_record
                add_num += 1
            distance_matrix.at[ri, ci] = np.nan

        if return_pruned:
            return edge_matrix, distance_matrix
        else:
            return edge_matrix

    # get median haplotypes
    @staticmethod
    def calculate_median_vector(triplet, sp_sequence_df, max_n_cost, epsilon, n, gap, new_link_data, shared_median_haps, initial_lambda, n_cost, lock):

        median_haps = pd.DataFrame(columns=['Haplotype'])
        temp_link_data = np.empty(shape=(0, 3))
        median_num = n

        for item in triplet:
            hap_in_order = item
            # check whether two haplotypes can be linked
            condition1 = any(hap_in_order[0]==np.array(new_link_data)[:, 0]) and any(hap_in_order[1]==np.array(new_link_data)[:, 1])
            condition2 = any(hap_in_order[0]==np.array(new_link_data)[:, 0]) and any(hap_in_order[2]==np.array(new_link_data)[:, 1])
            condition3 = any(hap_in_order[1]==np.array(new_link_data)[:, 0]) and any(hap_in_order[2]==np.array(new_link_data)[:, 1])
            if (condition1+condition2+condition3) < 2:
                continue
            # get alleles that are variable
            sub_seq = sp_sequence_df.loc[list(hap_in_order)]
            # get the median haplotype
            sub_seq_length = len(sub_seq.at[hap_in_order[0], 'Haplotype'])
            median_seq = ''.join([Counter(sub_seq['Haplotype'].map(lambda x: x[i])).most_common(1)[0][0] for i in range(sub_seq_length)])
            # check whether median sequence exist
            if median_seq in sp_sequence_df['Haplotype'].tolist():
                continue
            if not median_haps.empty:
                if median_seq in shared_median_haps:
                    continue
            # calculate the distance between median sequence and haplotypes
            if gap == 'missing':
                median_distance = [sum (l1 != l2 for l1, l2 in zip(i, median_seq) if l1 not in self.skip_letters and l2 not in self.skip_letters) for i in sub_seq['Haplotype']]
                if 0 in median_distance:
                    median_distance = [1 if dist==0 else dist for dist in median_distance]
            else:
                median_distance = [sum (l1 != l2 for l1, l2 in zip(i, median_seq)) for i in sub_seq['Haplotype']]
            connection_cost = sum(median_distance)
            # update n cost
            lock.acquire()
            n_cost.value += 1
            if n_cost.value > max_n_cost:
                print('[Y-LineageTracker] Program stopped since there are too many connnections')
                sys.exit()
            if connection_cost < initial_lambda.value:
                # update initial lambda
                initial_lambda.value = connection_cost
            if connection_cost <= initial_lambda.value + epsilon:
                hap_to_add1 = hap_in_order
                hap_to_add2 = [median_num]*3
                step_to_add = median_distance
                link_data_to_add = np.array([hap_to_add1, hap_to_add2, step_to_add]).T
                temp_link_data = np.concatenate([temp_link_data, link_data_to_add])
                median_haps.loc[median_num] = median_seq
                median_num += 1
                if median_seq not in shared_median_haps:
                    shared_median_haps.append(median_seq)
            lock.release()

        return median_haps, temp_link_data

    # construct network from distance matrix
    def _calculate_network(self, distance_matrix, sp_sequence_df, tree_type, epsilon):

        # initialization
        n = sp_sequence_df.index.size
        haplotype_length = len(sp_sequence_df.at[0, 'Haplotype'])
        median_num = n
        MST, pruned_distance_matrix = self._calculate_minimum_spanding_tree(n ,distance_matrix)

        if n < 3 or tree_type == 'msn':
            self.logger.warning('There are less than 3 types of haplotypes')
            alt_data = np.empty(shape=(0, 3))
            return MST, alt_data

        MST_step = MST[:, 2]
        lower_limit = min(MST_step)
        upper_limit = max(MST_step) + epsilon
        new_link_data = []
        for i in pruned_distance_matrix.columns:
            for j in pruned_distance_matrix.index:
                step = pruned_distance_matrix.at[j, i]
                if (not np.isnan(step)) and step > lower_limit and step <= upper_limit:
                    new_link_data.append([i, j, step])
        if new_link_data:
            new_link_data = np.concatenate([MST, np.array(new_link_data)])
        else:
            new_link_data = MST

        start_n_cost = 0
        start_initial_lambda = float('inf')
        tripleted = []

        while True:

            # get triplet for calculation
            triplet_new = list(itertools.combinations(range(n), 3))
            triplet = [i for i in triplet_new if i not in tripleted]
            tripleted.extend(triplet)

            process_num = 10
            pool = multiprocessing.Pool(processes=process_num)

            # global variables to be shared
            manager = multiprocessing.Manager()
            n_cost = manager.Value('i', start_n_cost)
            initial_lambda  = manager.Value('f', start_initial_lambda)
            shared_new_link_data = manager.list(new_link_data)
            shared_median_haps = manager.list([])
            LOCK = manager.Lock()

            # split triplet to sections
            triplet_interval = np.linspace(0, len(triplet), process_num, dtype=int)
            #triplet_sections = [triplet[triplet_interval[n]: triplet_interval[n+1]] if n == 0 else triplet[(triplet_interval[n]+1): triplet_interval[n+1]] for n in range(process_num-1)]
            triplet_sections = [triplet[triplet_interval[n]: triplet_interval[n+1]] for n in range(process_num-1)]

            # multiprocess triplet data
            pt = partial(self.calculate_median_vector, # function
                         sp_sequence_df=sp_sequence_df, # variable for read
                         max_n_cost=40000, # variable for read
                         epsilon=epsilon, # variable for read
                         n=n, # variable for read
                         gap=self.gap, # variable for read
                         new_link_data=new_link_data, # variable for writing
                         shared_median_haps=shared_median_haps, # variable for writing
                         initial_lambda=initial_lambda, # variable for writing
                         n_cost=n_cost, # variable for writing
                         lock=LOCK) # process lock
            median_haps_list, temp_link_data_list = zip(*pool.map(pt, triplet_sections))

            pool.close()
            pool.join()

            # intergrate data
            median_haps = pd.concat(median_haps_list)
            temp_link_data = np.concatenate(temp_link_data_list)
            if median_haps.empty:
                self.logger.info('[Y-LineageTracker] Get all median haplotypes')
                break
            else:
                # code for median vectors
                start_num = sp_sequence_df.index.size
                end_num = start_num + median_haps.index.size
                median_haps.index = range(start_num, end_num)
                temp_link_data[:, 2] = [val for val in range(start_num, end_num) for _ in (0, 1, 2)]
                # add to haplotype information
                sp_sequence_df = pd.concat([sp_sequence_df, median_haps])
                new_link_data = np.concatenate([new_link_data, temp_link_data])
                # update for iteration
                start_n_cost = n_cost.value
                start_initial_lambda = initial_lambda.value

        sp_sequence_df = sp_sequence_df.drop_duplicates().reset_index()

        self.logger.info('[Y-LineageTracker] Calculating distance matrix...')
        # calculate and add median haplotype to distance matrix
        hap_num = 0
        for hap1_name, hap1 in zip(sp_sequence_df.index, sp_sequence_df['Haplotype']):
            hap_num += 1
            for hap2_name, hap2 in zip(sp_sequence_df.index[hap_num:], sp_sequence_df['Haplotype'][hap_num:]):
                if self.gap == 'missing':
                    distance = sum(l1 != l2 for l1, l2 in zip(hap1, hap2) if l1 not in self.skip_letters and l2 not in self.skip_letters)
                    if distance == 0:
                        distance = 1
                else:
                    distance = sum(l1 != l2 for l1, l2 in zip(hap1, hap2))
                distance_matrix.at[hap2_name, hap1_name] = distance


        N = sp_sequence_df.index.size
        dnet = distance_matrix.copy().applymap(lambda x: 0 if not np.isnan(x) else x)
        forest = np.linspace(0, N-1, N)
        network_data = np.empty(shape=(0, 3))
        alt_data = np.empty(shape=(0, 3))
        steps = sorted(list(set(filter(lambda x: x==x, distance_matrix.to_numpy().reshape(-1)))))

        edge_num = 0
        add_new_data = False

        self.logger.info('[Y-LineageTracker] Constructing relationship between nodes')
        while len(set(forest)) > 1:

            step_num = steps[edge_num]
            step_df = distance_matrix[distance_matrix==step_num]
            step_count = step_df.notnull().sum().sum()
            for i in range(step_count):
                ci, ri = np.unravel_index(np.nanargmin(step_df.T.values), step_df.T.shape)
                step_df.at[ri, ci] = np.nan
                row_record = forest[ri]
                col_record = forest[ci]
                if row_record != col_record:

                    for col in np.where(forest==col_record)[0]:
                        if ci > col:
                            num_col = dnet.at[ci, col]
                        elif ci < col:
                            num_col = dnet.at[col, ci]
                        else:
                            num_col = 0
                        for row in np.where(forest==row_record)[0]:
                            if ri > row:
                                num_row = dnet.at[ri, row]
                            elif ri < row:
                                num_row = dnet.at[row, ri]
                            else:
                                num_row = 0
                            dnet_num = num_col + num_row + step_num
                            dnet.at[row, col] = dnet_num
                    forest[forest==col_record] = row_record
                    add_new_data = True
                else:
                    if step_num < dnet.at[ri, ci]:
                        add_new_data = True
                        # update dnet
                        cols = np.append(network_data[network_data[:, 1]==ci, 0], network_data[network_data[:, 0]==ci, 1])
                        rows = np.append(network_data[network_data[:, 1]==ri, 0], network_data[network_data[:, 0]==ri, 1])
                        for col in np.append(ci, cols).astype(int):
                            if ci > col:
                                num_col = dnet.at[ci, col]
                            elif ci < col:
                                num_col = dnet.at[col, ci]
                            else:
                                num_col = 0
                            for row in np.append(ri, rows).astype(int):
                                if col == row:
                                    continue
                                else:
                                    if ri > row:
                                        num_row = dnet.at[ri, row]
                                    elif ri < row:
                                        num_row = dnet.at[row, ri]
                                    else:
                                        num_row = 0
                                    dnet_num = num_col + num_row + step_num
                                    if row > col:
                                        if dnet.at[row, col] > dnet_num:
                                            dnet.at[row, col] = dnet_num
                                    elif row < col:
                                        if dnet.at[col, row] > dnet_num:
                                            dnet.at[col, row] = dnet_num
                if add_new_data:
                    network_to_add = np.array([[ci, ri, step_num]])
                    network_data = np.concatenate([network_data, network_to_add])
                    add_new_data = False
            edge_num += 1
            if edge_num > 5000:
                self.logger.error('Program stopped since there are too many connnections')
                sys.exit()

        network_data = network_data.astype(int)
        net_MST = self._calculate_minimum_spanding_tree(N ,distance_matrix, False)
        net_MST = net_MST.astype(int)

        if len(network_data) > len(net_MST):
            alt_data = np.array([i for i in network_data if i[:2].tolist() not in net_MST[:, :2].tolist()])
            network_data = np.array([i for i in network_data if i[:2].tolist() in net_MST[:, :2].tolist()])
        else:
            alt_data = np.array([])

        self.logger.info('[Y-LineageTracker] Calculating median haplotypes finished')

        return network_data, alt_data, sp_sequence_df

    # prune network for visualization
    def _get_network_for_plot(self, network_data, alt_data, hap_num, alt_type):

        # specify network data according to type of alternative link
        if alt_data.size != 0:
            alt_type = alt_type
            pruned_network_data = network_data.copy()
            max_number = np.max(pruned_network_data[:, :2]) + 1
            while True:
                to_be_deleted_row = []
                for i in range(hap_num, max_number):
                    if np.count_nonzero(pruned_network_data[:, :2]==i) == 1:
                        row_index = np.where(pruned_network_data[:, :2]==i)[0]
                        to_be_deleted_row.append(row_index)
                pruned_network_data = np.delete(pruned_network_data, to_be_deleted_row, axis=0)
                if len(to_be_deleted_row) == 0:
                    break
            # get network data for visulization
            if alt_type == 'hide':
                network_data = pruned_network_data
        else:
            alt_type = 'hide'
            pruned_network_data = network_data.copy()

        self.net_info.write('#HapLinks:\n')
        [self.net_info.write('\t'.join(i)+'\n') for i in pruned_network_data.astype(str)]
        self.net_info.close()
        node_num = len(np.unique(network_data[:, :2]))

        return network_data, node_num

    # get mutation number
    def _get_mutations(self, network_line, sp_sequence_df, STR_code_pos):

        seq1 = sp_sequence_df.at[network_line[0], 'Haplotype']
        seq2 = sp_sequence_df.at[network_line[1], 'Haplotype']
        mutation = [STR_code_pos[n] for n, s in enumerate(zip(seq1, seq2)) if s[0] != s[1]]

        return mutation

    # write and output fdi file
    def _get_fdi_file(self, network_data, torso_num, node_num, sequence_df, sp_sequence_df, sample_sp_code, pos, hap_size, hap_freq, STR_code_pos, pops):

        fdi_file = open(self.path+'.fdi', 'w')
        medians = sorted(list(set(set(range(len(sample_sp_code), sp_sequence_df.index.size)) & set(pos.keys()))))

        # head info
        center_x,  center_y = np.mean(np.array(list(pos.values())).T, axis=1)
        fdi_file.write('SHOW_NODENAMES;TRUE\n'\
                       'NODES_PROPORTIONAL;TRUE\n'\
                       'SHOW_ONLY_TORSO;FALSE\n'\
                       'SHOW_MEDIANS;FALSE\n'\
                       'SHOW_MEDIAN_NAMES;TRUE\n'\
                       'SHOW_CHARACTERS;FALSE\n'\
                       'SHOW_LINES;FALSE\n'\
                       'MIN_CIRC_RADIUS;4\n'\
                       'MAX_CIRC_RADIUS;50\n'\
                       'MED_RADIUS;3\n'\
                       '\n'\
                       'SPECIAL_FILE_TYPE;FALSE\n'\
                       'DIAGRAM_CENTER;%d;%d\n'\
                       'IMAGE_CENTER;%d;%d\n'\
                       % (center_x, center_y, center_x, center_y))

        # taxon info
        from FilesIO import set_color_num
        from matplotlib import colors
        rgb_colors = [list(map(lambda x: round(x*255), c)) for c in set_color_num(len(pops))]
        colors = [i[0]+i[1]*256+i[2]*256*256 for i in rgb_colors]
        fdi_file.write('NUMBER_OF_TAXA;%d\n' % (sequence_df.index.size+len(medians))) # including median nodes
        plot_num = 0
        for name in sequence_df.index:
            if name in sample_sp_code.keys():
                size = hap_size[plot_num]
                freq = hap_freq[plot_num]
                pos_x, pos_y = list(map(round, pos[plot_num]))
                active = 'TRUE'
                plot_num += 1
            else:
                size = 1
                pos_x, pos_y = 0, 0
                active = 'FALSE'
                freq = None
            taxon_info = 'TAXON_NAME;%s;'\
                         'TAXON_NAME_HEIGHT;6;'\
                         'TAXON_NAME_COLOR;0;'\
                         'TAXON_FREQUENCY;%d;'\
                         'TAXON_ORIG_FREQUENCY;1;'\
                         'TAXON_GEOGRAPHY;;'\
                         'TAXON_PHENOTYPE;;'\
                         'TAXON_LINEAGE;;'\
                         'TAXON_GROUP1;Group1;;'\
                         'TAXON_GROUP2;Group2;;'\
                         'TAXON_GROUP3;Group3;;'\
                         'TAXON_X;%d;'\
                         'TAXON_Y;%d;'\
                         % (name, size, pos_x, pos_y)
            if freq:
                pie_info = ''
                colored_pie_num = 0
                for pie_num, pie_freq in enumerate(freq):
                    if pie_freq == 0:
                        continue
                    else:
                        colored_pie_num += 1
                        pie_color_code = colors[pie_num]
                        pie_info += 'TAXON_COLOR_PIE%d;%d;'\
                                    'TAXON_PIE_FREQUENCY%d;%d;'\
                                    'TAXON_STYLE_PIE%d;SOLID;'\
                                    % (colored_pie_num, pie_color_code, colored_pie_num, pie_freq, colored_pie_num)
            else:
                pie_info = 'TAXON_COLOR_PIE1;65535;'\
                           'TAXON_PIE_FREQUENCY1;1;'\
                           'TAXON_STYLE_PIE1;SOLID;'\

            line_info = 'TAXON_LINE_WIDTH;1;'\
                        'TAXON_LINE_COLOR;0;'\
                        'TAXON_LINE_STYLE;SOLID;'\
                        'TAXON_ACTIVE;%s\n'\
                        % (active)

            fdi_file.write(taxon_info + pie_info + line_info)

        # median info
        for n, m in enumerate(medians):
            pos_x, pos_y = list(map(round, pos[m]))
            median_info = 'TAXON_NAME;mv%d;'\
                          'TAXON_NAME_HEIGHT;6;'\
                          'TAXON_NAME_COLOR;0;'\
                          'TAXON_FREQUENCY;0;'\
                          'TAXON_ORIG_FREQUENCY;0;'\
                          'TAXON_GEOGRAPHY;;'\
                          'TAXON_PHENOTYPE;;'\
                          'TAXON_LINEAGE;;'\
                          'TAXON_GROUP1;Group1;;'\
                          'TAXON_GROUP2;Group2;;'\
                          'TAXON_GROUP3;Group3;;'\
                          'TAXON_X;%d;'\
                          'TAXON_Y;%d;'\
                          'TAXON_COLOR_PIE1;255;'\
                          'TAXON_PIE_FREQUENCY1;0;'\
                          'TAXON_STYLE_PIE1;SOLID;'\
                          'TAXON_LINE_WIDTH;1;'\
                          'TAXON_LINE_COLOR;0;'\
                          'TAXON_LINE_STYLE;'\
                          'SOLID;TAXON_ACTIVE;TRUE\n'\
                          % (n, pos_x, pos_y)
            fdi_file.write(median_info)

        fdi_file.write('\n')
        fdi_file.write('NUMBER_OF_LINKS;%d\n' % len(network_data))
        for n, link in enumerate(network_data):
            if n < torso_num:
                in_torso = 'TRUE'
            else:
                in_torso = 'FALSE'
            link_names = [list(sample_sp_code.keys())[name_code] if name_code < len(sample_sp_code) else 'mv'+str(medians.index(name_code)) for name_code in link[:2]]
            link_info = 'LINK_TAXON1;%s;'\
                        'LINK_TAXON2;%s;'\
                        'LINK_IN_TORSO;%s;'\
                        'LINK_TORSO_PART;1;'\
                        'NUMBER_OF_MUTATIONS;%d;'\
                        % (link_names[0], link_names[1], in_torso ,link[2])
            mutation_data = self._get_mutations(link, sp_sequence_df, STR_code_pos)
            mutation_info = ''
            for mutation_n, mutation in enumerate(mutation_data):
                mutation_info += 'MUTATION%d_NAME;%s;'\
                                 'MUTATION%d_NAME_HEIGHT;6;'\
                                 'MUTATION%d_NAME_COLOR;255;'\
                                 'LINK_LINE_WIDTH;1;'\
                                 'LINK_LINE_STYLE;SOLID;'\
                                 'LINK_LINE_COLOR;8421504;'\
                                 'LINK_ACTIVE;TRUE'\
                                 % (mutation_n+1, mutation, mutation_n+1, mutation_n+1)
            fdi_file.write(link_info+mutation_info+'\n')

        # other info
        # median nodes not included
        fdi_file.write('\n')
        other_info = 'NUMBER_OF_HELP_POINT_LINKS;0\n'\
                     '\n'\
                     'LONG_EDGES;0\n'\
                     'TOTAL_TAXA;%d\n'\
                     'TOTAL_HAPLOTYPES;%d\n'\
                     % (sequence_df.index.size, len(sample_sp_code))
        fdi_file.write(other_info)

        # equal info
        equal_info = ''
        equal_num = 0
        for sample, hap in sample_sp_code.items():
            equal_samples = sequence_df[sequence_df['Haplotype']==hap].index.tolist()
            for equal in equal_samples:
                if equal != sample:
                    equal_num += 1
                    equal_info += 'EQUIVALENT_TAXA;%s;%s\n' % (sample, equal)
        fdi_file.write('NUMBER_EQUIVALENT_TAXA;%d\n' % equal_num)
        fdi_file.write(equal_info)


        # end info
        end_info = 'COLOR_SCHEME;SCHEME_ACTIVE;FALSE;SCHEME_TYPE;Phenotype;\n'\
                   'COLOR_SCHEME;SCHEME_ACTIVE;FALSE;SCHEME_TYPE;Geography;\n'\
                   'COLOR_SCHEME;SCHEME_ACTIVE;FALSE;SCHEME_TYPE;Lineage;\n'\
                   'COLOR_SCHEME;SCHEME_ACTIVE;FALSE;SCHEME_TYPE;Group1;\n'\
                   'COLOR_SCHEME;SCHEME_ACTIVE;FALSE;SCHEME_TYPE;Group2;\n'\
                   'COLOR_SCHEME;SCHEME_ACTIVE;FALSE;SCHEME_TYPE;Group3;\n'
        fdi_file.write(end_info)

        fdi_file.close()

    # summarize parameters for figure
    def _get_figure_parameters(self, sequence_df, population_data, network_data, alt_data, node_num, original_network_size, alt_type, euqal_size, label):

        haplotypes = sorted(list(set(sequence_df['Haplotype'])))
        # calculate frequency of each haplotypes
        if euqal_size:
            hap_size = [1]*len(haplotypes)
        else:
            hap_size = [sequence_df['Haplotype'].tolist().count(hap) for hap in haplotypes]
            hap_size_per = [size / sequence_df.index.size for size in hap_size ]


        if isinstance(population_data, pd.DataFrame):
            pops = sorted(list(set(population_data[label])))
            # calculate haplotype frequency in each population
            hap_pop_data = pd.DataFrame(index=range(len(haplotypes)), columns=pops)
            for pop in pops:
                pop_inds = population_data[population_data[label]==pop]['SampleID'].tolist()
                pop_data = sequence_df.loc[pop_inds]
                hap_pop_data[pop] = [pop_data['Haplotype'].tolist().count(hap) for hap in haplotypes]
            hap_freq = [hap_pop_data.loc[i].tolist() for i in hap_pop_data.index]
            hap_freq_per = [(hap_pop_data.loc[i]/sum(hap_pop_data.loc[i])).round(6).tolist() for i in hap_pop_data.index]
        else:
            hap_freq_per = None
            pops = None

        # set length of each node
        length_data = np.empty(shape=(0, 3))
        num = np.max(network_data[:, :2])+1
        to_be_deleted_row = []
        for row_index, i in enumerate(network_data):
            step = i[2]
            if step > 1:
                start = i[0]
                end = i[1]
                for i in range(step):
                    if i == 0:
                        length = [start, num, 1]
                    elif i == step-1:
                        length = [num, end, 1]
                    else:
                        length = [num, num+1, 1]
                        num += 1
                    length_data = np.concatenate((length_data, np.array([length])))
                num += 1
                to_be_deleted_row.append(row_index)
                original_network_size -= 1
        network_data = np.delete(network_data, to_be_deleted_row, axis=0)
        length_data = length_data.astype(int)

        # set figure size and network class
        import warnings
        warnings.filterwarnings('ignore')
        font_files = font_manager.findSystemFonts(fontpaths=[os.path.split(os.path.realpath(__file__))[0] + '/sans-serif'])
        font_list = font_manager.createFontList(font_files)
        font_manager.fontManager.ttflist.extend(font_list)
        plt.rcParams['font.family']= 'sans-serif'
        plt.rcParams['font.sans-serif'] = ['Arial'] # font
        plt.rcParams['legend.handlelength'] = 1
        plt.rcParams['legend.handleheight'] = 1
        fig_len = len(length_data)/3
        fig, ax = plt.subplots(figsize=(fig_len, fig_len))
        trans = ax.transData.transform
        trans2 = fig.transFigure.inverted().transform
        DG = nx.DiGraph()

        # construct network by adding nodes and edges
        style_list = []
        for i, j in enumerate(network_data):
            if i < original_network_size:
                DG.add_edge(j[0], j[1])
                style_list.append('solid')
            else:
                DG.add_edge(j[0], j[1])
                if alt_type != 'dashed':
                    style_list.append('dashed')
                else:
                    style_list.append('solid')
        # construct network by adding node of length
        for i in length_data:
            DG.add_edge(i[0], i[1])
            style_list.append('solid')

        try:
            import pygraphviz as pg
            pos = nx.nx_agraph.graphviz_layout(DG, prog='neato')
        except ImportError:
            pre_pos = nx.spring_layout(DG)
            pos = {}
            for i in pre_pos.keys():
                pos[i] = trans(pre_pos[i])

        return length_data, DG, pos, hap_size, hap_freq, hap_size_per, hap_freq_per, style_list, pops, trans, trans2, fig_len

    # plot a tree and replace nodes to pies
    def plot_network(self, length_data, DG, pos, node_num, median_size_rate, hap_size_per, hap_freq_per, style_list, pops, trans, trans2, fig_len, label):


        import warnings
        warnings.filterwarnings('ignore')

        # set figure parameters
        # mdian haplotype size
        base_size = 5*20
        median_size = base_size * median_size_rate
        # all node size
        hap_num = len(hap_size_per)
        max_size = max(hap_size_per)
        size_rate = base_size/max_size
        size = [hap_size_per[i]*size_rate if i < hap_num else median_size for i in DG.nodes][:node_num]
        size = size + [0.2] * length_data.size

        try:
            # color and line type
            if hap_freq_per:
                from matplotlib.patches import Patch
                from FilesIO import set_color_num
                nx.draw(DG, pos=pos, arrows=False, node_color='black', node_size=min(size), style=style_list)
                # set colors
                colors = set_color_num(len(pops))
                legend_elements = []
                for color, pop in zip(colors , pops):
                    legend_elements.append(Patch(facecolor=color, edgecolor='black', label=pop))
                plt.legend(handles=legend_elements, frameon=False, bbox_to_anchor=(-0.04, 0.5), prop={'size': fig_len})
                pie_size = 0.025
                move = pie_size/2.0
                for i in DG:
                    if i < hap_num:
                        xx, yy = trans(pos[i])
                        xa, ya = trans2((xx,yy))
                        pie_ax = plt.axes([xa-move, ya-move, pie_size, pie_size])
                        pie_ax.set_aspect('equal')
                        fracs = hap_freq_per[i]
                        if 1 in fracs:
                            one_colors= [colors[fracs.index(1)]]
                            one_fracs = [1]
                            pie_ax.pie(one_fracs, colors=one_colors, radius=(1/(max_size/hap_size_per[i]))**0.5, wedgeprops={'edgecolor':'black','linewidth': 0.5, 'linestyle': 'solid'})
                        else:
                            pie_ax.pie(fracs, colors=colors, radius=(1/(max_size/hap_size_per[i]))**0.5, wedgeprops={'edgecolor':'black','linewidth': 0.5, 'linestyle': 'solid'})
            else:
                colors = ['yellow' if i < hap_num else 'black' for i in DG.nodes] # color
                nx.draw(DG, pos=pos, arrows=False, node_color=colors, node_size=size, style=style_list, linewidths=1)
                node_ax = plt.gca() # to get the current axis
                node_ax.collections[0].set_edgecolor('black')
                node_ax.collections[0].set_linewidth(0.5)

            # output
            if label:
                plt.savefig(self.path+'.net.fig.%s.pdf' % label, dpi=100, bbox_inches='tight')
            else:
                plt.savefig(self.path+'.net.fig.pdf', dpi=100, bbox_inches='tight')
            plt.close()
        except AttributeError:
            print('[Y-LineageTracker] Cannot generate figure since matplotlib.cbook version bugs')

    # main function of network analysis
    def calculate_and_output(self, haplotype_data, population_data, arguments, label):

        # convert data to 0-1 code and calculate distance matrix
        sequence_df, STR_code_pos = self._convert_allel_to_binary(haplotype_data, arguments.filter)
        distance_matrix, sp_sequence_df, sample_sp_code = self._calculate_distance_matrix(sequence_df)
        # calculate network
        self.logger.info('[Y-LineageTracker] Calculating median haplotypes')
        network_data, alt_data, sp_sequence_df2 = self._calculate_network(distance_matrix, sp_sequence_df, arguments.treetype, arguments.search)
        plot_network_data, node_num = self._get_network_for_plot(network_data, alt_data, sp_sequence_df.index.size, arguments.alt)

        # visulization
        self.logger.info('[Y-LineageTracker] Start to plot network')

        if label != 'Group':
            length_data, DG, pos, hap_size, hap_freq, hap_size_per, hap_freq_per, style_list, pops, trans, trans2, fig_len = self._get_figure_parameters(sequence_df, population_data, plot_network_data, alt_data, node_num, network_data.size, arguments.alt, arguments.euqalsize, label)
            if arguments.fdi:
                self._get_fdi_file(plot_network_data, plot_network_data.size, node_num, sequence_df, sp_sequence_df2, sample_sp_code, pos, hap_size, hap_freq, STR_code_pos, pops)
            self.plot_network(length_data, DG, pos, node_num, arguments.mediansize, hap_size_per, hap_freq_per, style_list, pops, trans, trans2, fig_len, label)
        else:
            label = ['Population', 'Group']
            for i in label:
                length_data, DG, pos, hap_size, hap_freq, hap_size_per, hap_freq_per, style_list, pops, trans, trans2, fig_len = self._get_figure_parameters(sequence_df, population_data, plot_network_data, alt_data, node_num, network_data.size, arguments.alt, arguments.euqalsize, label)
                if arguments.fdi:
                    self._get_fdi_file(plot_network_data, plot_network_data.size, node_num, sequence_df, sp_sequence_df2, sample_sp_code, pos, hap_size, hap_freq, STR_code_pos, pops)
                self.plot_network(length_data, DG, pos, node_num, arguments.mediansize, hap_size_per, hap_freq_per, style_list, pops, trans, trans2, fig_len, label)

        self.logger.info('[Y-LineageTracker] Network analysis finished')



def main():

    start = time.perf_counter()
    arguments = net_parser()

    from FilesIO import check_seq_matrix_input, check_hap_input, set_haplotype_index, check_population, check_overlap, check_pop_header, time_count
    # check input files
    input, format = check_seq_matrix_input(arguments.seq, arguments.format, arguments.matrix)

    # check haplotype data
    haplotype_data = check_hap_input(input, 'haplotype', format)

    # check filter argument
    check_misisng_rate(arguments.filter)

    from FilesIO import get_out_path
    # get path of output files
    path = get_out_path(arguments.output, 'Net')
    log_file = path + '.NetLog.log'

    # set log file
    set_log(arguments, log_file)

    # check if there is a population file
    if arguments.population:
        population_data = check_population(arguments.population)
        population_data = check_pop_header(population_data)
        haplotype_data, population_data = check_overlap(haplotype_data, population_data)
        label = 'Population'
        if 'Group' in population_data.columns:
            label = 'Group'
    else:
        population_data = None
        label = ''

    # set index of haplotype data
    haplotype_data = set_haplotype_index(haplotype_data, format)

    # calculate network and visualize result
    net = HaploNetwork(arguments.gap, path)
    net.calculate_and_output(haplotype_data, population_data, arguments, label)

    time_count(start)


if __name__ == '__main__':
    main()



import time
import logging
import argparse

from ClassifyHaplogroup import MatchVariant


'''
Description:
This module is used for genotyping Y-STRs from BAM or VCF indels.
Different Y-STR panels are supported in this command and can be selected by the –-panel option.
The accuracy and genotyping rate depend on the quality of input data (such as reads coverage of BAM files).
'''


def genostr_parser():

    parser = argparse.ArgumentParser('genostr', description='(c) Y-LineageTracker: Y-STR genotyping')
    # function used for Y-STR genotyping
    parser.add_argument('genostr',
                        help='Genotype Y-STR from BAM or VCF indels.')
    # required, input file, BAM or VCF
    input = parser.add_mutually_exclusive_group(required=True)
    input.add_argument('--vcf',
                        type=str,
                        action='store',
                        help='input: The file name of input data used for Y-STR genotyping.')
    input.add_argument('--bam',
                        type=str,
                        action='append',
                        help='input: The file name of input data used for Y-STR genotyping.')
    # optional, provide samples to be analyzed
    parser.add_argument('-s', '--sample',
                        required=False,
                        type=str,
                        action='store',
                        help='Sample: A file containing sample individuals only.')
    # optional, version of reference genome: 37 or 38
    parser.add_argument('-b', '--build',
                        required=False,
                        type=int,
                        action='store',
                        default=37,
                        choices=[37, 38],
                        help='build: The build version of reference genome used for Y-STR genotyping. (37 or 38).')
    # optional, filter females by misisng rate
    parser.add_argument('--filter',
                        required=False,
                        type=str,
                        action='store',
                        nargs='?',
                        const='0.75',
                        help='filter: Filter female individuals by setting cutoff of missing rate, default is 0.75. If the missing rate value is not set, the program will automatically filter female individuals by the built-in algorithm.')
    # optional, the set of Y-STR loci used for genotyping
    parser.add_argument('--panel',
                        required=False,
                        type=str,
                        default='named',
                        choices=['all', 'named', 'minimal', 'ppy', 'pp23', 'yf'],
                        help='panel: The set of Y-STR loci for genotyping. These sets are referenced from commonly used STR multiplex kits:\n'
                             'all: all the possible Y-STR loci\n'
                             'named: all the possible named Y-STR loci\n'
                             'minimal: minimal panel for Y-STR loci\n'
                             'ppy: PowerPlex Y panel for Y-STR loci\n'
                             'pp23: PowerPlex 23 panel for Y-STR loci\n'
                             'Yf: Yfiler panel for Y-STR loci')
    # optional, the prefix of output
    parser.add_argument('-o', '--output',
                        required=False,
                        type=str,
                        action='store',
                        help='output: The prefix of output files.')

    args = parser.parse_args()

    return args


# print program information and write to log file
def set_log(input, log_file, format, cutoff, args_log):

    logger = logging.getLogger()
    logger.setLevel(level=logging.INFO)

    handler = logging.FileHandler(log_file, mode='w')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] - [%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    logger.addHandler(handler)
    logger.addHandler(console)

    log_info = ['[Y-LineageTracker] [GenotypeSTR]',
                '[Y-LineageTracker] Run Date: ' + time.asctime(time.localtime(time.time())),
                '[Y-LineageTracker] Input File: %s' % input,
                '[Y-LineageTracker] File Format: %s' % format,
                '[Y-LineageTracker] Reference Build: GRCh%d' % args_log.build,
                '[Y-LineageTracker] STR Panel: %s' % args_log.panel]

    if args_log.sample:
        log_info.append('[Y-LineageTracker] List of samples provided')
    else:
        if not cutoff is None:
            if cutoff:
                log_info.append('[Y-LineageTracker] Missing rate cutoff: %.2f' % cutoff)
            else:
                log_info.append('[Y-LineageTracker] Filter female individuals automatically')

    print('\n')
    for i in log_info:
        logger.info(i)

# names of output files
def check_str_output(output):

    from FilesIO import get_out_path
    path = get_out_path(output, 'Genostr')
    out_file = path + '.str.matrix'
    log_file = path + '.StrGenotype.log'

    return out_file, log_file


def main():

    start = time.perf_counter()
    arguments = genostr_parser()

    from FilesIO import check_cutoff, check_BAM_VCF_input, check_samples, time_count
    # check input files
    input, format = check_BAM_VCF_input(arguments.vcf, arguments.bam)

    # check cutoff: cutoff is a float between 0 and 1, else False
    cutoff = check_cutoff(arguments.filter, format)

    # set of output files
    out_file, log_file = check_str_output(arguments.output)

    # set log file
    set_log(input, log_file, format, cutoff, arguments)

    # check whether provide sample list
    samples = check_samples(arguments.sample)

    # filter samples
    from ProcessData import ReadFilterData
    read_filter = ReadFilterData.ReadFilterData(input, samples)

    # if BAM, genotype Y-STR from ReadFilterData module
    if format == 'bam':
        result = read_filter.read_bam(arguments.panel, arguments.build, 'str')[0]
    # if VCF filter samples from ReadFilterData, then genotype Y-STRs from MatchVariant
    else:
        data_info, inds = read_filter.read_filter_variant(cutoff, format)
        # genotype Y-STR
        match = MatchVariant.MatchVariant(arguments.build)
        result = match.match_variant_str(data_info, inds, arguments.panel)

    # output result in matrix format
    result.index.name = 'SampleID'
    result.to_csv(out_file,
                  sep='\t',
                  index=True)

    time_count(start)


if __name__ == '__main__':
    main()



import sys

version = '1.2.16'
help_info = '\n' \
            '[Y-LineageTracker] a framework to fully analyze human Y-chromosome sequencing data from paternal level\n' \
            '\n' \
            'Version: ' + version + '\n' \
            'Usage: LineageTracker <subcommand> [options]\n' \
            '\n' \
            '[Subcommands]\n' \
            '--NRY haplogroup analysis\n' \
            '    classify    Classify NRY haplogroups and construct lineages from BAM or VCF.\n' \
            '    cluster     Perform NRY haplogroup clustering analysis.\n' \
            '    phylo       Perform Phylogenetic analysis for NRY haplogroups.\n' \
            '\n' \
            '--Y-STR and Y-haplotype analysis\n' \
            '    genostr     Genotype Y-STR from BAM or VCF indels.\n' \
            '    net         Perform network analysis from Y haplotype data.\n' \
            '    stat        Perform statistical analysis from Y haplotype data.\n' \
            '\n' \
            '--Time estimation\n' \
            '    time        Estimate divergence time of NRY haplogroups.\n' \
            '    tmrca       Estimate TMRCA of Y haplotypes.\n' \
            '\n' \
            'For more details about subcommands, type: LineageTracker <subcommand> -h/--help\n' \
            '\n' \
            '[General Options]\n' \
            '    -h/--help       Print help massages.\n' \
            '    -v/--version    Display the version of Y-LineageTracker you are using.\n'


def main():

    if len(sys.argv) > 1:
        # command for testing
        if sys.argv[1] == 'test':
            from Test import TestProgram
            TestProgram.main()
        # commands for analysis
        elif sys.argv[1] == 'classify':
            import ClassifyHaplogroup
            ClassifyHaplogroup.main()
        elif sys.argv[1] == 'cluster':
            import ClusterHaplogroup
            ClusterHaplogroup.main()
        elif sys.argv[1] == 'phylo':
            import PhyloHaplogroup
            PhyloHaplogroup.main()
        elif sys.argv[1] == 'genostr':
            import GenotypeSTR
            GenotypeSTR.main()
        elif sys.argv[1] == 'net':
            import HaploNetwork
            HaploNetwork.main()
        elif sys.argv[1] == 'stat':
            import StatisticsAnalysis
            StatisticsAnalysis.main()
        elif sys.argv[1] == 'time':
            import EstimateTime
            EstimateTime.main()
        elif sys.argv[1] == 'tmrca':
            import CommonAncestor
            CommonAncestor.main()
        elif sys.argv[1] == 'test':
            from Test import TestProgram
        # command for version information
        elif sys.argv[1] in ['version', '--version', '-v']:
            print('[Y-LineageTracker] version: %s' % version)
        # command for help information
        elif sys.argv[1] in ['help', '--help', '-h']:
            print(help_info)
        else:
            print(help_info)
    else:
        print(help_info)

if __name__ == '__main__':
    main()



import os
import sys
import subprocess

from setuptools import Command, find_packages, setup
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info
from LineageTracker import __VERSION__


MIN_PY_VER = (3, 6)
if sys.version_info[:2] < MIN_PY_VER:
    sys.stderr.write(
        ("Biopython requires Python %i.%i or later. " % MIN_PY_VER)
        + ("Python %d.%d detected.\n" % sys.version_info[:2])
    )
    sys.exit(1)


required = ['pandas',
            'numpy',
            'scikit-learn',
            'matplotlib',
            'ete3',
            'networkx',
            'BioPython',
            'pysam',
            'pyqt5',
            'adjustText']
request = ['pygraphviz']


def get_virtualenv_path():
    '''Used to work out path to install compiled binaries to.'''

    if hasattr(sys, 'real_prefix'):
        return sys.prefix

    if hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix:
        return sys.prefix

    if 'conda' in sys.prefix:
        return sys.prefix

    return None


def compile_and_install_software():

    '''Used the subprocess module to compile/install the C software.'''
    src_path = './LineageTracker/src/'

    # install the software (into the virtualenv bin dir if present)
    try:
        subprocess.check_call('make', cwd=src_path, shell=True)
    except:
        print('make command is required for installation')

    try:
        import platform
        system = platform.system()
        if platform == 'Windows':
            subprocess.check_call('pip install pygraphviz', shell=True)
        else:
            subprocess.check_call('pip install --install-option="--include-path=/usr/local/include/" --install-option="--library-path=/usr/local/lib/" pygraphviz', shell=True)
    except:
        print('Failed to install pygraphviz')

class CustomInstall(install):
    '''Custom handler for the 'install' command.'''

    def run(self):
        install.run(self)
        compile_and_install_software()
        super().run()

class CustomDevelopCommand(develop):

    def run(self):
        develop.run(self)
        compile_and_install_software()


class CustomEggInfoCommand(egg_info):

    def run(self):
        egg_info.run(self)
        compile_and_install_software()


class TestProgram(Command):

    description = 'run to test Y-LineageTracker'
    user_options = [('cmd', None, 'test all commands'),]

    def initialize_options(self):
        '''Set default values for options'''
        self.cmd = 'all'

    def finalize_options(self):
        '''Post-process options'''
        print('[Y-LineageTracker] [Test] Testing started')

    def run(self):
      '''Run command'''
      import importlib
      for i in required:
          found = importlib.util.find_spec(i)
      if not found:
          print('[Y-LineageTracker] [Test] Package [%s] is required for testing')
          sys.exit()

      current_dir = os.path.split(os.path.realpath(__file__))[0]
      sys.path.insert(0, current_dir+ '/Test')
      from Test import TestProgram
      if self.cmd == 'all':
          TestProgram.main()


setup(name='Y-LineageTracker',
      version=__VERSION__,
      author='Hao Chen',
      author_email='chenhao@picb.ac.cn',
      description='Y-LineageTracker: A framework to fully analyze human Y-chromosome sequencing data',
      license='GPL-3.0',
      keywords=['Bioinformatics', 'Computational biology', 'Population genetics',
                'Genomics', 'Y chromosome', 'haplogroup', 'Y-STR'],
      url='http://www.picb.ac.cn/PGG/resource.php',
      packages=['LineageTracker',
                'LineageTracker.ProcessData',
                'LineageTracker.GetLineage',
                'LineageTracker.Test'],
      package_data={'LineageTracker': ['Data/*', 'sans-serif/*', 'src/*', 'Test/TestData/*', 'Config.ini']},
      entry_points={'console_scripts':
                    ['LineageTracker = LineageTracker.RunLineagerTracker:main']},
      install_requires=required,
      classifiers=['Development Status :: 3 - Alpha',
                   'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Science/Research',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: C',
                   'Topic :: Scientific/Engineering :: Bio-Informatics',
                   'Topic :: Scientific/Engineering :: Visualization'],
      zip_safe=False,
      cmdclass={'test': TestProgram,
                'install': CustomInstall,
                'develop': CustomDevelopCommand,
                'egg_info': CustomEggInfoCommand},
      python_requires=">=%i.%i" % MIN_PY_VER)



import os
import sys
import time
import logging
import argparse
import warnings
import itertools

import pandas as pd


class CommonData(object):
    '''
    This class is used to read common data for specific analysis
    '''

    def __init__(self, snp_only=False):

        warnings.filterwarnings('ignore')
        self.data_path = os.path.split(os.path.realpath(__file__))[0] + '/Data/'

    def read_region_info(self):

        region_info = pd.read_csv(self.data_path + 'Y_chromosome_region.bed',
                                  header=0,
                                  sep='\t',
                                  index_col=False,
                                  encoding='utf-8')

        return region_info

    def read_ref_info(self, snp_only, approximate):

        ref_info = pd.read_csv(self.data_path + 'ISOGG_haplogroup_info.csv',
                               header=0,
                               sep=',',
                               encoding='utf-8',
                               dtype='object')
        if not approximate:
            ref_info = ref_info[ref_info['ApproximateInfo']!='Approximate']
        if snp_only:
            ref_info = ref_info[ref_info['Variant'].map(lambda x: len(x) == 1)]
        ref_ind_info = ref_info[ref_info['ReferenceInfo']=='Ref']

        return ref_info, ref_ind_info


    def read_pop_info(self):

        pop_info = pd.read_csv(self.data_path + 'ISOGG_population_index.csv',
                               header=0,
                               sep=',',
                               encoding='utf-8')

        return pop_info

    def read_str_info(self):

        str_info = pd.read_csv(self.data_path + 'Y_STR_panel.csv',
                               header=0,
                               sep=',',
                               encoding='utf-8')

        return str_info

    def read_str_prediction_panel(self):

        Y_STR_precidtion_panel = pd.read_csv(self.data_path + 'Y_STR_precidtion_panel.csv',
                                             header=0,
                                             sep=',',
                                             encoding='utf-8')

        return Y_STR_precidtion_panel

    def read_prior_calibration(self):

        prior_calibration = pd.read_csv(self.data_path + 'haplogroup_calibration.csv',
                                        header=0,
                                        index_col=0,
                                        sep=',',
                                        encoding='utf-8')

        return prior_calibration

    def read_tree_info(self):

        tree_file = open(self.data_path + 'haplogroup_tree_2018.txt', 'r+')
        tree_info = list(map(lambda x: x.replace('\n', ''), tree_file.readlines()))
        tree_file.close()

        return tree_info


# get output path
def get_out_path(output, name):

    if output:
        if not os.path.isdir(output):
            if os.path.isdir(os.path.split(output)[0]) or os.path.split(output)[0] == '':
                path = output
            else:
                raise FileNotFoundError('No such directory: %s' % os.path.split(output)[0])
        else:
            raise FileNotFoundError('%s is directory' % output)
    else:
        path = os.getcwd() + '/' + name

    return path


# get STR panel
def get_str_from_panel(panel):

    common_data = CommonData()
    str_info = common_data.read_str_info()
    # choose which kit will be used to genotype Y-STR
    if panel != 'all':
        str_info = str_info[str_info['STR'].map(lambda x: x != '.')]
    if panel != 'named':
        str_info = str_info[str_info['Kit'].map(lambda x: not pd.isna(x))]
        if panel == 'minimal':
            str_info = str_info[str_info['Kit'].map(lambda x: 'Minimal' in x.split(', '))]
        elif panel == 'ppy':
            str_info = str_info[str_info['Kit'].map(lambda x: 'PowerPlex_Y' in x.split(', '))]
        elif panel == 'pp23':
            str_info = str_info[str_info['Kit'].map(lambda x: 'PowerPlex_Y23' in x.split(', '))]
        elif panel == 'yf':
            str_info = str_info[str_info['Kit'].map(lambda x: 'Yfiler' in x.split(', '))]

    return str_info


# check whether set cutoff of filter
def check_cutoff(filter, format):

    if format == 'bam':
        if filter:
            logger = logging.getLogger()
            logger.warning('[Y-LineageTracker] [Warning] filter option only used for VCF' % len(overlap))
        cutoff = None
    else:
        if filter == None:
            cutoff = None
        elif filter == 'auto':
            cutoff = False
        else:
            try:
                cutoff = float(filter)
            except:
                raise argparse.ArgumentTypeError('Cutoff of missing rate should be a float between 0 and 1')
            if cutoff < 0 or cutoff > 1:
                raise argparse.ArgumentTypeError('Cutoff of missing rate should be a float between 0 and 1')

    return cutoff


# check bam or vcf input
def check_BAM_VCF_input(vcf, bam):

    if vcf:
        input = vcf
        compressed = check_compress(input)
        if compressed:
            format = 'gzvcf'
        else:
            format = 'vcf'
    else:
        input = ', '.join(bam)
        format = 'bam'

    return input, format


# check sequence alignment file or matrix file
def check_seq_matrix_input(seq, pre_format, matrix):

    if seq:
        input = seq
        if pre_format:
            format = pre_format
        else:
            raise argparse.ArgumentTypeError('Fomrat should be specified for sequence file by --seq-format')
    else:
        input = matrix
        format = 'matrix'

    return input, format


# check whther file is compressed
def check_compress(input_file):

    if os.path.isfile(input_file):
        try:
            open(input_file).read()
            compressed = False
        except UnicodeDecodeError:
            compressed = True
    else:
        raise FileNotFoundError('%s not found, please check your file name' % input_file)
        sys.exit()

    return compressed


# check whether sample file is provided
def check_samples(male_file):

    if male_file:
        try:
            with open(male_file, 'r') as f:
                males = f.read().splitlines()
        except:
            raise FileNotFoundError('Unable to open male file, please check your file name')
    else:
        males = []

    return males


# check header and value type of input data
def check_hap_input(input_file, type, format='matrix'):

    if type == 'haplotype' and format != 'matrix':
            from ProcessData import ConvertData
            convert = ConvertData.ConvertData(input_file)
            input_data = convert.data_convert(format, 'matrix')
    else:
        input_data = pd.read_csv(input_file,
                                 sep='\t',
                                 header=0,
                                 encoding='utf-8')
    input_data = input_data.apply(lambda x: x.str.strip() if x.dtype == "object" else x) # strip whitespace

    if type == 'haplogroup':
        if input_data.columns.size != 2 or input_data.columns.tolist() != ['SampleID', 'Haplogroup']:
            raise KeyError('Header of haplogroup file not correct, please check the header is: SampleID, Haplogroup')
    elif type == 'haplotype':
        if input_data.columns.size < 2 or input_data.columns.tolist()[0] != 'SampleID':
            raise KeyError('Header of haplotype file not correct, please check the header is: SampleID, Haplotype Names')

    return input_data


# check population file
def check_population(population_file):

    try:
        population_data = pd.read_csv(population_file,
                                      sep='\t',
                                      header=0,
                                      encoding='utf-8')
        population_data = population_data.apply(lambda x: x.str.strip() if x.dtype == "object" else x)
    except:
        raise FileNotFoundError('%s not found, please check your file name' % population_file)
    condition_a = population_data.columns.size == 2 and population_data.columns[0:2].tolist() == ['SampleID', 'Population']
    condition_b = population_data.columns.size == 3 and population_data.columns[0:3].tolist() == ['SampleID', 'Population', 'Group']
    condition_c = population_data.columns.size == 3 and population_data.columns[0:3].tolist() == ['SampleID', 'Population', 'Color']
    condition_d = population_data.columns.size == 3 and population_data.columns[0:3].tolist() == ['SampleID', 'Population', 'Shape']
    condition_e = population_data.columns.size == 4 and population_data.columns[0:4].tolist() == ['SampleID', 'Population', 'Group', 'Color']
    condition_f = population_data.columns.size == 4 and population_data.columns[0:4].tolist() == ['SampleID', 'Population', 'Group', 'Shape']
    condition_g = population_data.columns.size == 4 and population_data.columns[0:4].tolist() == ['SampleID', 'Population', 'Color', 'Shape']
    condition_h = population_data.columns.size == 5 and population_data.columns[0:5].tolist() == ['SampleID', 'Population', 'Group', 'Color', 'Shape']
    if not (condition_a or condition_b or condition_c or condition_d or condition_e or condition_f or condition_g or condition_h):
        raise KeyError('Header of population file not correct, please check the header is: SampleID, Population or SampleID, Population, Group')

    return population_data


# check population header
def check_pop_header(population_data):

    if 'Group' in population_data.columns.tolist():
        population_data = population_data[['SampleID', 'Population', 'Group']]
    else:
        population_data = population_data[['SampleID', 'Population']]

    return population_data


# set SampleID as index of haplotype data
def set_haplotype_index(haplotype_data, format):

    from GetConfig import getConfig
    config = getConfig()
    skip_letters = config.get('Statistics', 'SkipLetters').split(',')

    haplotype_data = haplotype_data.set_index(['SampleID'], drop=True)
    if format == 'matrix':
        if haplotype_data.applymap(lambda x: isinstance(x, int) or x.isdigit() or x in skip_letters).sum().sum() != haplotype_data.size:
            raise KeyError('Unexpected value in input haplotype matrix data')

    return haplotype_data


# check whther individuals are overlapped in two files
def check_overlap(sample_data, population_data):

    logger = logging.getLogger()

    # sort and drop duplicated individuals
    sample_data = sample_data.sort_values(by='SampleID').drop_duplicates().reset_index(drop=True)
    population_data = population_data.sort_values(by='SampleID').drop_duplicates().reset_index(drop=True)
    # whther samples are same of two files
    if sample_data.index.size == population_data.index.size:
        if all(sample_data['SampleID']==population_data['SampleID']):
            pass
        else:
            # only keep matched indvisuals
            TF = sample_data['SampleID']==population_data['SampleID']
            sample_data = sample_data[TF]
            population_data = population_data[TF]
            overlap_num = sum(sample_data['SampleID']==population_data['SampleID'])
            logger.info('[Y-LineageTracker] [Warning] Only %d individuals overlaped in haplogroup file and population file' % len(overlap_num))
    else:
        overlap = sorted(list(set(sample_data['SampleID'])&set(population_data['SampleID'])))
        if len(overlap) > 0:
            # only keep matched indvisuals
            sample_data = sample_data[sample_data['SampleID'].map(lambda x: x in overlap)]
            population_data = population_data[population_data['SampleID'].map(lambda x: x in overlap)]
            logger.info('[Y-LineageTracker] [Warning] Only %d individuals overlaped in haplogroup file and population file' % len(overlap))
        else:
            # No matched indvisuals
            logger.error('[Y-LineageTracker] Format error or There are no overlaped individuals in haplogroup/haplotype file and population file')
            sys.exit()
    population_data = population_data.sort_values(by=population_data.columns.tolist()[::-1][1:]).reset_index(drop=True)

    return sample_data, population_data


# set default colors
def set_color_num(num, get_cmap=False):

    from matplotlib import cm
    from GetConfig import getConfig

    config = getConfig()
    color_num = num
    mim_color_num = config.getint('PlotParameters', 'MinColorNum')
    median_color_num = config.getint('PlotParameters', 'MedianColorNum')
    max_color_num = config.getint('PlotParameters', 'MaxColorNum')
    if color_num <= mim_color_num:
        cmap = cm.get_cmap('Set1').colors
    elif color_num > mim_color_num and color_num <= median_color_num:
        cmap = cm.get_cmap('tab20').colors
    elif color_num > median_color_num and color_num <= max_color_num:
        cmap = cm.get_cmap('tab20').colors + cm.get_cmap('tab20b').colors
    colors = list(cmap)[:color_num]

    if get_cmap:
        return colors, cmap
    else:
        return colors


# haplogroup normalization
def haplogroup_normalize(haplogroup):

    upper_trunk = config.get('HaplogroupTree', 'UpperTrunk')

    if not haplogroup[0].isupper():
        haplogroup = haplogroup[0].upper()+haplogroup[1:]
    if not len(haplogroup) == 1:
        if haplogroup.upper() in upper_trunk:
            haplogroup = haplogroup.upper()
        else:
            if not haplogroup[1:].islower():
                haplogroup = haplogroup[0] + haplogroup[1:].lower()

    return haplogroup


# get end time to calculate total run time
def time_count(start, mid=False):

    logger = logging.getLogger()

    interval = time.perf_counter() - start
    if interval > 60:
        minute = int(interval/60)
        second = interval - minute * 60

    if mid:
        if interval > 60:
            logger.info('[Y-LineageTracker] Time: %dmin%.2fs' % (minute, second))
        else:
            logger.info('[Y-LineageTracker] Time: %.2fs' % (interval))
    else:
        if interval > 60:
            logger.info('[Y-LineageTracker] Total run time: %dmin%.2fs' % (minute, second))
        else:
            logger.info('[Y-LineageTracker] Total run time: %.2fs' % (interval))
